/**
 * \file ipoltitle.c
 * \brief A command line program that generates IPOL video title screen
 * \author Pascal Getreuer <getreuer@gmail.com>
 *
 * This program generates a title screen of an animated IPOL logo with a
 * specified article title and DOI.  The frame resolution and frame rate
 * can be configured through the command line options.  The output of this
 * program is a set of frame images: frame-0001.png, frame-0002.png, 
 * frame-0003.png, etc.
 *
 * To create the frames, it is required that a command processor is 
 * available (for the system command) and that pdflatex and ImageMagick 
 * convert are on the path.  The program works by first creating the frame
 * as a LaTeX document, then calling pdflatex and convert to compile and
 * rasterize to PNG.
 *
 * The frames can be assembled into a video with ffmpeg,
@code
  ffmpeg -r 29.97 -i frame-%04d.png -vcodec libx264 -crf 22 title.mp4
@endcode
 * or as an animated GIF, MNG, or MPEG with ImageMagick convert,
@code 
   convert -delay 3 frame-*.png -layers optimize title.gif
@endcode
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "point3d.h"

/* Command to invoke pdflatex */
#define PDFLATEX_COMMAND        "pdflatex -halt-on-error -interaction=nonstopmode"
/* Command to invoke ImageMagick convert */
#define CONVERT_COMMAND         "convert"

#define FADE_DURATION           0.8
#define BEFORE_ANIM_DURATION    0.7
#define ANIM_DURATION           1.2
#define AFTER_ANIM_DURATION     1.7
/* Total duration of the title screen in seconds */
#define TOTAL_DURATION          \
    (BEFORE_ANIM_DURATION + ANIM_DURATION + AFTER_ANIM_DURATION)

/* Parameter defaults */
#define DEFAULT_FRAME_WIDTH     720
#define DEFAULT_FRAME_HEIGHT    540
#define DEFAULT_FPS             29.97
#define DEFAULT_ANTIALIAS       2

/* IPOL logo colors in HTML hex notation */
#define C_LOGO_1                "231F20"
#define C_LOGO_2                "5140A0"
#define C_LOGO_3                "836CBD"
#define C_LOGO_4                "B09ED6"
#define C_LOGO_5                "C9BCE4"
#define C_LOGO_TEXT             "505050"

/* LaTeX document paper width in inches */
#define PAPER_WIDTH             2.4

/** \brief Print program usage message */
void print_usage()
{
    puts("ipoltitle, IPOL video title screen generator, P. Getreuer 2012\n");
    puts("Usage: make_title [options] -t <article_title> -d <article_doi>\n");

    puts("This program generates a  title screen of an animated IPOL logo with");
    puts("a specified article title and DOI.  The frame resolution and frame");
    puts("rate can be configured through the command line options.  The");
    puts("output of this program is a set of frame images: f0001.png,");
    puts("f0002.png, f0003.png, etc.\n");

    puts("Options:");
    puts("  -t  article title");
    puts("  -d  article DOI");
    printf("  -w  frame width (default %d)\n", DEFAULT_FRAME_WIDTH);
    printf("  -h  frame height (default %d)\n", DEFAULT_FRAME_HEIGHT);
    printf("  -f  frames per second (default %g)\n", DEFAULT_FPS);
    printf("  -a  antialiasing factor (default %d)\n", DEFAULT_ANTIALIAS);
    puts("  -o  output directory (default is current directory)");
    puts("  -l  generate only the last frame\n");
}

/** \brief 3D square in the logo */
typedef struct
{
    point3d points[4];
    double alpha;
} square3d;

/** \brief Video generation parameters */
typedef struct 
{    
    /* These parameters are constant throughout the video */
    const char *output_dir;     /**< output directory                   */
    const char *article_title;  /**< article title as LaTeX code        */  
    const char *article_doi;    /**< article DOI as LaTeX code          */
    int frame_width;            /**< frame width in pixels              */
    int frame_height;           /**< frame height in pixels             */
    double frames_per_second;   /**< frames per second                  */
    int num_frames;             /**< total number of frames             */
    int antialias_factor;       /**< rasterization antialiasing factor  */
    int only_last_frame;        /**< if nonzero, only create last frame */
    
    /* These parameters vary from one frame to the next */
    int current_frame;          /**< Index (base 1) of current frame    */
    char output_stem[256];      /**< filestem of the current frame      */
    char tex_filename[256];     /**< filename for LaTeX document        */    
    square3d squares[5];        /**< 3D squares in the logo             */
    double global_alpha;        /**< Global alpha for fade in effect    */
} params;

void update_3d_model(params *p);
void write_latex(const params *p);
void call_pdflatex(params *p);
void call_convert(params *p);


/** \brief Make title main routine */
void make_title(params *p)
{
    if(!system(NULL))   /* Check whether system() calls work */
        fprintf(stderr, "Warning: No command processor available\n");
        
    p->num_frames = (int)(TOTAL_DURATION * p->frames_per_second + 0.5);
    p->current_frame = (p->only_last_frame) ? p->num_frames : 1;
    
    printf("Number of frames: %d\n", p->num_frames);
    
    while(p->current_frame <= p->num_frames)
    {
        printf("Frame %d of %d\n", p->current_frame, p->num_frames);
        sprintf(p->output_stem, "%s/f%04d", p->output_dir, p->current_frame);
        sprintf(p->tex_filename, "%s.tex", p->output_stem);

        update_3d_model(p); /* Update the animated 3D model */
        write_latex(p);     /* Write LaTeX file for the current frame */
        call_pdflatex(p);   /* Compile with pdflatex */
        call_convert(p);    /* Rasterize to PNG */
        
        p->current_frame++;
    }
}

/** \brief Print formatted error message and exit */
void xerror(const char *message, ...)
{
    va_list args;
    fputs("Error: ", stderr);
    va_start(args, message);
    vfprintf(stderr, message, args);
    va_end(args);
    exit(1);
}

/** \brief S-curve function */
double scurve(double x, double crossover, double radius)
{
    if(x <= crossover - radius)
        return 0;
    else if(x >= crossover + radius)
        return 1;
    
    x = (x - crossover) / radius;
    return (3 - x*x)*x/4 + 0.5;
}

/** \brief Compute position and alphas of the 3D model for the current frame */
void update_3d_model(params *p)
{
    static const square3d square_ref = 
        {{{-0.5, 0, -0.5}, {0.5, 0, -0.5}, {0.5, 0, 0.5}, {-0.5, 0, 0.5}}, 0};
    double theta = 45;
    double phi = -45;
    mat4x4 T;
    double t = (p->current_frame - 1)/p->frames_per_second;
    int i, j;
    
    p->global_alpha = scurve(t, 0.5*FADE_DURATION, 0.5*FADE_DURATION);
    T = mat4x4_compose(mat4x4_rot_pitch(phi), mat4x4_rot_roll(theta));
    
    for(i = 0; i < 5; i++)
    {
        double s, y;
        
        if(t <= BEFORE_ANIM_DURATION)
            s = 1;
        else if(t > BEFORE_ANIM_DURATION + ANIM_DURATION)
            s = 0;
        else
            s = 1 - (t - BEFORE_ANIM_DURATION) / ANIM_DURATION;        
        
        if(i == 0)
        {
            y = 0;
            p->squares[i].alpha = 1.0;
        }
        else
        {
            y = 0.22 + 3.5*s - (4 - i)*0.5;
            
            if(y < 0.22*i)
                y = 0.22*i;
            
            p->squares[i].alpha = 0.8 * scurve(1 - s, (i + 1)/7.0, 1.75/7.0);
        }
        
        y -= 0.4;
        
        for(j = 0; j < 4; j++)
            p->squares[i].points[j] = 
                mat4x4_transform_point3d(
                    mat4x4_compose(T, mat4x4_translate(0, y, 0)),
                    square_ref.points[j]);
    }
}

void write_latex_preamble(FILE *file, const params *p)
{
    fprintf(file,   /* Header and packages */
        "\\documentclass[11pt]{article}\n"
        "\\usepackage[papersize={%.6fin,%.6fin},margin=0in]{geometry}\n"
        "\\usepackage{xcolor,tikz}\n",
        PAPER_WIDTH, (PAPER_WIDTH * p->frame_height) / p->frame_width);
    fprintf(file,   /* Define colors */
        "\\definecolor{clogo1}{HTML}{" C_LOGO_1 "}\n"
        "\\definecolor{clogo2}{HTML}{" C_LOGO_2 "}\n"
        "\\definecolor{clogo3}{HTML}{" C_LOGO_3 "}\n"
        "\\definecolor{clogo4}{HTML}{" C_LOGO_4 "}\n"
        "\\definecolor{clogo5}{HTML}{" C_LOGO_5 "}\n"
        "\\definecolor{clogotext}{HTML}{" C_LOGO_TEXT "}\n");
    fprintf(file,   /* More preamble */
        "\\renewcommand{\\familydefault}{\\sfdefault}\n"
        "\\parskip=0pt\n"
        "\\parindent=0pt\n"
        "\\pagestyle{empty}\n");
}

void write_tikzpicture(FILE *file, const params *p)
{
    int i;
    
    /* Begin tikzpicture */
    fprintf(file, 
        "\\begin{tikzpicture}%%\n");
    fprintf(file,   /* Set tikzpicture bounding box */
        "\\path [use as bounding box,clip]"
        " (0,0) rectangle (\\textwidth,1.3);\n");
    fprintf(file,   /* Draw horizontal line below logo */
        "\\draw [gray!%d] (-0.1in,0) -- (2.5in,0);\n",
        (int)(p->global_alpha * 40 + 0.5));
    
    /* Begin a scope in which to draw the logo */
    fprintf(file,
        "\\begin{scope}[scale=0.5,xshift=15mm,yshift=12mm]\n");
    
    for(i = 4; i >= 0; i--)
        fprintf(file,   /* Draw square */
            "\\fill [clogo%d,opacity=%.2f]"
            " (%.4f,%.4f) -- (%.4f,%.4f) --"
            " (%.4f,%.4f) -- (%.4f,%.4f);\n",
            i + 1, p->squares[i].alpha * p->global_alpha,
            p->squares[i].points[0].x, p->squares[i].points[0].z,
            p->squares[i].points[1].x, p->squares[i].points[1].z,
            p->squares[i].points[2].x, p->squares[i].points[2].z,
            p->squares[i].points[3].x, p->squares[i].points[3].z);
    
    fprintf(file,   /* Add "Image Processing On Line" and link */
        "\\draw (0.8,-0.08) node [right,clogotext,opacity=%.2f]"
        " {Image Processing On Line};\n"
        "\\draw (0.8,-0.78) node [right,clogotext,opacity=%.2f]"
        " {\\scriptsize www.ipol.im};\n",
        p->global_alpha, p->global_alpha);
    
    /* End of tikzpicture*/
    fprintf(file, 
        "\\end{scope}\n"
        "\\end{tikzpicture}%%\n");
}

/** \brief Write LaTeX document for current frame */
void write_latex(const params *p)
{
    FILE *file = NULL;
    
    if(!(file = fopen(p->tex_filename, "wt")))
        xerror("Unable to write to \"%s\"\n", p->tex_filename);
    
    fprintf(file, "%% IPOL title animation, frame %d of %d\n",
        p->current_frame, p->num_frames);
    write_latex_preamble(file, p);
    
    /* Begin LaTeX document */
    fprintf(file,
        "\\begin{document}%%\n"
        "\\centering%%\n");
    
    write_tikzpicture(file, p);
    
    fprintf(file, "\\vspace{4ex}%%\n\n");
    fprintf(file, "\\textcolor{black!%d}{",
        (int)(100.0 * p->global_alpha + 0.5));
    fprintf(file, "{\\Large %s} \\\\\n", p->article_title);
    fprintf(file, "{\\scriptsize %s}}\n", p->article_doi);
    
    /* End of LaTeX document */
    fprintf(file,
        "\\end{document}\n");    
    fclose(file);
}

/** \brief Call pdflatex to compile current LaTeX frame */
void call_pdflatex(params *p) 
{
    char command[1024];
    int status;
    
    sprintf(command, PDFLATEX_COMMAND " -output-directory=%s %s", 
        p->output_dir, p->tex_filename);
    puts(command);
    
    if((status = system(command)))
        fprintf(stderr, 
            "Warning: pdflatex exited with nonzero status %d\n", status);
}

/** \brief Call convert to rasterize PDF to PNG */
void call_convert(params *p)
{
    char command[1024];
    int status;
    
    sprintf(command, 
        CONVERT_COMMAND " -density %.6f %s.pdf"
        " -background white -flatten -resize %dx%d -depth 8 %s.png",
        (p->antialias_factor * p->frame_width) / ((double)PAPER_WIDTH), 
        p->output_stem,
        p->frame_width, p->frame_height, p->output_stem);
    puts(command);
    
    if((status = system(command)))
        fprintf(stderr, 
            "Warning: convert exited with nonzero status %d\n", status);
}

/** \brief Parse parameters from command line arguments */
int parse_arguments(params *p, int argc, char *argv[])
{
    const char *default_output_dir = ".";
    const char *default_article_title = "Article Title";
    const char *default_article_doi = "http://dx.doi.org/10.5201/ipol.2012.x";
    char *option_string;
    char option_char;
    int i;

    if(argc < 2)
    {   /* Program was called without arguments */
        print_usage();
        return 0;
    }

    /* Set parameter defaults */
    p->output_dir = default_output_dir;
    p->article_title = default_article_title;
    p->article_doi = default_article_doi;
    p->frame_width = DEFAULT_FRAME_WIDTH;
    p->frame_height = DEFAULT_FRAME_HEIGHT;
    p->frames_per_second = DEFAULT_FPS;
    p->antialias_factor = DEFAULT_ANTIALIAS;
    p->only_last_frame = 0;
    
    for(i = 1; i < argc;)
    {
        if(argv[i] && argv[i][0] == '-')
        {
            if((option_char = argv[i][1]) == 0)
                xerror("Invalid parameter format.\n");

            if(argv[i][2])
                option_string = &argv[i][2];
            else if(++i < argc)
                option_string = argv[i];
            else if(option_char != 'l')
                xerror("Invalid parameter format.\n");
            
            switch(option_char)
            {
            case 't':
                p->article_title = option_string;
                break;
            case 'd':
                p->article_doi = option_string;
                break;
            case 'w':
                p->frame_width = atoi(option_string);
                break;
            case 'h':
                p->frame_height = atoi(option_string);
                break;
            case 'f':
                p->frames_per_second = atof(option_string);
                break;
            case 'a':
                p->antialias_factor = atoi(option_string);
                break;
            case 'o':
                p->output_dir = option_string;
                break;
            case 'l':
                p->only_last_frame = 1;
                i--;
                break;
            
            case '-':
                print_usage();
                return 0;
            default:
                if(isprint(option_char))
                    xerror("Unknown option \"-%c\".\n", option_char);
                else
                    xerror("Unknown option.\n");
                return 0;
            }

            i++;
        }
    }
    
    return 1;
}

int main(int argc, char **argv)
{
    params p;
    
    if(!parse_arguments(&p, argc, argv))
        return 1;
    
    make_title(&p);
    return 0;
}
